/**
	This program illustrates a simple way to generate certain 
	textures procedurally.

	See http://www.code-spot.co.za/... for more details.

	You can alter the following parameters to generate different 
	types of textures: 
	
		PROB_MATRIX (If you change the matrix's size, 
			make sure to change WINDOW_WIDTH as well.)
			This matrix is used to select appropriate 
			pixels to blend in.

		HALF_ITERATION_COUNT Half the number of iterations.
*/

#include <iostream>

#include "pngwriter.h"
#include <cstdlib>

#include <vector>
#include <cassert>

using std::cout;
using std::endl;


#define IM_SIZE 500
#define WINDOW_WIDTH 5

#define PROB_MATRIX  {   \
		0, 1, 0, 1, 0,   \
		0, 0, 1, 0, 0,   \
		0, 0, 1, 0, 0,   \
		0, 0, 1, 0, 0,   \
		0, 0, 1, 0, 0    \
	}

#define HALF_ITERATION_COUNT 50

/**
	Creates a random double between 0 and 1.
*/
double random()
{
	return (rand() + 1) / (double) RAND_MAX;
}

/**
	Generates a grid with white noise (double values between 0 and 1).
*/
void white_noise(double * grid , int width, int height )
{
	for (int i = 0; i < width; i++)
	{
		for (int j = 0; j < height; j++)
		{
			grid[i + width*j] = random();
		}
	}
}

/**
	Writes an image to disk.
*/
void write_image(double * grid, int width, int height, const char * fname)
{
	pngwriter png(width, height, 0, fname);
	
	for (int i = 0; i < width; i++)
	{
		for (int j = 0; j <  height; j++)
		{			
			int n = i + j * width;
			png.plot(i+1, j+1, grid[n], grid[n], grid[n]);
		}
	}

	png.close();
}

/**
	Blends two grayscale values in the ratio t.
*/
double mix(double col1, double col2, double t)
{
	return col1 * (1.0 - t) + col2 * t;
}

/**
	Takes an image, and blurs it according to the algorithm. The result is put into new_grid.
*/
void blur_once(double * old_grid, double * new_grid, int width, int height, double acc[], double t)
{
	for (int i = 0; i < width; i++)
	{
		for (int j = 0; j < height; j++)
		{
			double r = random();
			int k = 0;

			while ((acc[k] < r - 0.001) && (k < WINDOW_WIDTH*WINDOW_WIDTH))
			{
				k++;
			}

			int x = k % WINDOW_WIDTH - (WINDOW_WIDTH - 1) / 2;
			int y = k / WINDOW_WIDTH - (WINDOW_WIDTH - 1) / 2;

			int new_i = (width + i + x) % width;
			int new_j = (height + j + y) % height;	

			assert(new_i < width);
			assert(new_j < height);

			assert(new_i + new_j * width < width * height);

			new_grid[i + j * width] = mix(old_grid[i + j * width], old_grid[new_i + new_j * width], t);					
		}
	}
}

/**
	Stretches the range to [0 1].
*/
void normalise(double * grid, int width, int height)
{
	double min, max, scl;

	min = grid[0];
	max = grid[0];

	for(int i = 1; i < width*height; i++)
	{
		if (grid[i] < min)
		{
			min = grid[i];
		}
		else if (grid[i] > max)
		{
			max = grid[i];
		}
	}

	scl = 1.0 / (max - min);

	for(int i = 1; i < width*height; i++)
	{
		grid[i] -= min;
		grid[i] *= scl;
	}
}

int main()
{
	//seed the randomiser for experimentation to a constant value
	srand(40);

	double prob[] = PROB_MATRIX;
		
	for (int i = 0; i < WINDOW_WIDTH*WINDOW_WIDTH; i++)
	{
		prob[i] = random();

		cout << prob[i] << endl;
	}

	double sum = 0;

	for (int i = 0; i < WINDOW_WIDTH*WINDOW_WIDTH; i++)
		sum += prob[i];

	for (int i = 0; i < WINDOW_WIDTH*WINDOW_WIDTH; i++)
		prob[i] /= sum;

	double acc[WINDOW_WIDTH*WINDOW_WIDTH] = {0};

	acc[0] = prob[0];

	for (int i = 1; i < WINDOW_WIDTH*WINDOW_WIDTH; i++)
	{
		acc[i] = prob[i] + acc[i - 1];
	}

	double * grid1 = new double[IM_SIZE*IM_SIZE];
	double * grid2 = new double[IM_SIZE*IM_SIZE];

	white_noise(grid1, IM_SIZE, IM_SIZE);
	
	char name[125];

	for(int i = 0; i < HALF_ITERATION_COUNT; i++)
	{		
		printf("%d ", 2 * i);

		//we do this twice inside the loop, this way we can reuse the arrays.
		blur_once(grid1, grid2, IM_SIZE, IM_SIZE, acc, .5);
		blur_once(grid2, grid1, IM_SIZE, IM_SIZE, acc, .5);		
		
		normalise(grid1, IM_SIZE, IM_SIZE);
		sprintf(name, "test_im_%02d.png", (2 * i +1));
		write_image(grid1, IM_SIZE, IM_SIZE, name);
	}

	delete [] grid1;
	delete [] grid2;
	
	return 0;
}